// Copyright 2016 Google Inc. All rights reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

use std::cmp::{min,max};
use std::io::Write;

use serde_json::value::Value;

use xi_rope::rope::{Rope, LinesMetric, RopeInfo};
use xi_rope::delta::Delta;
use xi_rope::tree::Cursor;
use xi_rope::breaks::{Breaks, BreaksInfo, BreaksMetric, BreaksBaseMetric};
use xi_rope::interval::Interval;
use xi_rope::spans::Spans;
use xi_rope::find::{find, CaseMatching};

use tabs::{ViewIdentifier, DocumentCtx};
use styles;
use index_set::IndexSet;
use selection::{Affinity, Selection, SelRegion};
use movement::{Movement, selection_movement};

use linewrap;

const SCROLL_SLOP: usize = 2;
const BACKWARDS_FIND_CHUNK_SIZE: usize = 32_768;

#[derive(Default, Clone)]
pub struct Style {
    pub fg: u32,
    pub font_style: u8,  // same as syntect, 1 = bold, 2 = underline, 4 = italic
}

pub struct View {
    pub view_id: ViewIdentifier,

    /// The selection state for this view. Invariant: non-empty.
    selection: Selection,

    drag_state: Option<DragState>,

    first_line: usize,  // vertical scroll position
    height: usize,  // height of visible portion
    breaks: Option<Breaks>,
    wrap_col: usize,

    // Ranges of lines held by the line cache in the front-end that are considered
    // valid.
    // TODO: separate tracking of text, cursors, and styles
    valid_lines: IndexSet,

    // The selection state was updated.
    sel_dirty: bool,
    // The occurrences, which determine the highlights, have been updated.
    hls_dirty: bool,

    // TODO: much finer grained tracking of dirty state
    dirty: bool,

    /// Tracks whether or not the view has unsaved changes.
    pristine: bool,

    /// The currently active search string
    search_string: Option<String>,
    /// The case matching setting for the currently active search
    case_matching: CaseMatching,
    /// The set of all known find occurrences (highlights)
    occurrences: Option<Selection>,
    /// Set of ranges that have already been searched for the currently active search string
    valid_search: IndexSet,
}

/// State required to resolve a drag gesture into a selection.
struct DragState {
    /// All the selection regions other than the one being dragged.
    base_sel: Selection,

    /// Offset of the point where the drag started.
    offset: usize,

    /// Start of the region selected when drag was started (region is
    /// assumed to be forward).
    min: usize,

    /// End of the region selected when drag was started.
    max: usize,
}

impl View {
    pub fn new(view_id: &ViewIdentifier) -> View {
        let mut selection = Selection::new();
        selection.add_region(SelRegion {
            start: 0,
            end: 0,
            horiz: None,
            affinity: Affinity::default(),
        });
        View {
            view_id: view_id.to_owned(),
            selection: selection,
            drag_state: None,
            first_line: 0,
            height: 10,
            breaks: None,
            wrap_col: 0,
            valid_lines: IndexSet::new(),
            sel_dirty: true,
            hls_dirty: true,
            dirty: true,
            pristine: true,
            search_string: None,
            case_matching: CaseMatching::CaseInsensitive,
            occurrences: None,
            valid_search: IndexSet::new(),
        }
    }

    pub fn set_scroll(&mut self, first: usize, last: usize) {
        self.first_line = first;
        self.height = last - first;
    }

    pub fn scroll_height(&self) -> usize {
        self.height
    }

    pub fn scroll_to_cursor(&mut self, text: &Rope) {
        let end = self.sel_regions().last().unwrap().end;
        let (line, _) = self.offset_to_line_col(text, end);
        if line < self.first_line {
            self.first_line = line;
        } else if self.first_line + self.height <= line {
            self.first_line = line - (self.height - 1);
        }
    }

    /// Toggles a caret at the given offset.
    pub fn toggle_sel(&mut self, offset: usize) {
        self.sel_dirty = true;
        if !self.selection.regions_in_range(offset, offset).is_empty() {
            self.selection.delete_range(offset, offset);
            if !self.selection.is_empty() {
                self.drag_state = None;
                return;
            }
        }
        self.drag_state = Some(DragState {
            base_sel: self.selection.clone(),
            offset: offset,
            min: offset,
            max: offset,
        });
        let region = SelRegion {
            start: offset,
            end: offset,
            horiz: None,
            affinity: Affinity::default(),
        };
        self.selection.add_region(region);
    }

    /// Move the selection by the given movement. Return value is the offset of
    /// a point that should be scrolled into view.
    ///
    /// If `modify` is `true`, the selections are modified, otherwise the results
    /// of individual region movements become carets.
    pub fn do_move(&mut self, text: &Rope, movement: Movement, modify: bool)
        -> Option<usize>
    {
        self.drag_state = None;
        let new_sel = selection_movement(movement, &self.selection, self, text, modify);
        self.set_selection(text, new_sel)
    }

    /// Set the selection to a new value. Return value is the offset of a
    /// point that should be scrolled into view.
    pub fn set_selection(&mut self, text: &Rope, sel: Selection) -> Option<usize> {
        self.selection = sel;
        // We somewhat arbitrarily choose the last region for setting the old-style
        // selection state, and for scrolling it into view if needed. This choice can
        // likely be improved.
        let region = self.selection.last().unwrap().clone();
        self.sel_dirty = true;
        self.scroll_to_cursor(text);
        Some(region.end)
    }

    /// Select entire buffer.
    ///
    /// Note: unlike movement based selection, this does not scroll.
    pub fn select_all(&mut self, len: usize) {
        let mut selection = Selection::new();
        selection.add_region(SelRegion {
            start: 0,
            end: len,
            horiz: None,
            affinity: Default::default(),
        });
        self.selection = selection;
        self.sel_dirty = true;
    }

    /// Starts a drag operation.
    pub fn start_drag(&mut self, offset: usize, min: usize, max: usize) {
        let base_sel = Selection::new();
        self.drag_state = Some(DragState { base_sel, offset, min, max });
    }

    /// Does a drag gesture, setting the selection from a combination of the drag
    /// state and new offset.
    pub fn do_drag(&mut self, text: &Rope, offset: usize, affinity: Affinity) -> Option<usize> {
        let new_sel = self.drag_state.as_ref().map(|drag_state| {
            let mut sel = drag_state.base_sel.clone();
            // TODO: on double or triple click, quantize offset to requested granularity.
            let (start, end) = if offset < drag_state.offset {
                (drag_state.max, min(offset, drag_state.min))
            } else {
                (drag_state.min, max(offset, drag_state.max))
            };
            let horiz = None;
            sel.add_region(SelRegion { start, end, horiz, affinity });
            sel
        });
        new_sel.and_then(|new_sel| self.set_selection(text, new_sel))
    }

    /// Returns the regions of the current selection.
    pub fn sel_regions(&self) -> &[SelRegion] {
        &self.selection
    }

    /// Determines whether the offset is in any selection (counting carets and
    /// selection edges).
    pub fn is_point_in_selection(&self, offset: usize) -> bool {
        !self.selection.regions_in_range(offset, offset).is_empty()
    }

    // Render a single line, and advance cursors to next line.
    fn render_line<W: Write>(&self, tab_ctx: &DocumentCtx<W>, text: &Rope,
        start_of_line: &mut Cursor<RopeInfo>, soft_breaks: Option<&mut Cursor<BreaksInfo>>, style_spans: &Spans<Style>,
        line_num: usize) -> Value
    {
        let start_pos = start_of_line.pos();
        let pos = soft_breaks.map_or(start_of_line.next::<LinesMetric>(), |bc| {
            let pos = bc.next::<BreaksMetric>();
            // if using breaks update cursor
            if let Some(pos) = pos { start_of_line.set(pos) }
            pos
        }).unwrap_or(text.len());

        let l_str = text.slice_to_string(start_pos, pos);
        let mut cursors = Vec::new();
        let mut selections = Vec::new();
        for region in self.selection.regions_in_range(start_pos, pos) {
            // cursor
            let c = region.end;
            if (c > start_pos && c < pos) ||
                (!region.is_upstream() && c == start_pos) ||
                (region.is_upstream() && c == pos) ||
                (c == pos && c == text.len() && self.line_of_offset(text, c) == line_num)
            {
                cursors.push(c - start_pos);
            }

            // selection with interior
            let sel_start_ix = clamp(region.min(), start_pos, pos) - start_pos;
            let sel_end_ix = clamp(region.max(), start_pos, pos) - start_pos;
            if sel_end_ix > sel_start_ix {
                selections.push((sel_start_ix, sel_end_ix));
            }
        }

        let mut hls = Vec::new();
        if let Some(ref occurrences) = self.occurrences {
            for region in occurrences.regions_in_range(start_pos, pos) {
                let sel_start_ix = clamp(region.min(), start_pos, pos) - start_pos;
                let sel_end_ix = clamp(region.max(), start_pos, pos) - start_pos;
                if sel_end_ix > sel_start_ix {
                    hls.push((sel_start_ix, sel_end_ix));
                }
            }
        }

        let styles = self.render_styles(tab_ctx, start_pos, pos, &selections, &hls, style_spans);

        let mut result = json!({
            "text": &l_str,
            "styles": styles,
        });

        if !cursors.is_empty() {
            result["cursor"] = json!(cursors);
        }
        result
    }

    pub fn render_styles<W: Write>(&self, tab_ctx: &DocumentCtx<W>, start: usize, end: usize,
        sel: &[(usize, usize)], hls: &[(usize, usize)], style_spans: &Spans<Style>) -> Vec<isize>
    {
        let mut rendered_styles = Vec::new();
        let style_spans = style_spans.subseq(Interval::new_closed_open(start, end));

        let mut ix = 0;
        for &(sel_start, sel_end) in sel {
            rendered_styles.push((sel_start as isize) - ix);
            rendered_styles.push(sel_end as isize - sel_start as isize);
            rendered_styles.push(0);
            ix = sel_end as isize;
        }
        for &(sel_start, sel_end) in hls {
            rendered_styles.push((sel_start as isize) - ix);
            rendered_styles.push(sel_end as isize - sel_start as isize);
            rendered_styles.push(1);
            ix = sel_end as isize;
        }
        for (iv, style) in style_spans.iter() {
            // This conversion will move because we'll store style id's in the spans
            // data structure. But we're changing things one piece at a time.
            let new_style = styles::Style {
                fg: style.fg,
                bg: 0,
                weight: if (style.font_style & 1) != 0 { 700 } else { 400},
                underline: (style.font_style & 2) != 0,
                italic: (style.font_style & 4) != 0,
            };
            let style_id = tab_ctx.get_style_id(&new_style);
            rendered_styles.push((iv.start() as isize) - ix);
            rendered_styles.push(iv.end() as isize - iv.start() as isize);
            rendered_styles.push(style_id as isize);
            ix = iv.end() as isize;
        }
        rendered_styles
    }

    pub fn send_update<W: Write>(&mut self, text: &Rope, tab_ctx: &DocumentCtx<W>, style_spans: &Spans<Style>,
        first_line: usize, last_line: usize)
    {
        let dirty = self.dirty || self.sel_dirty;
        if dirty {
            self.valid_lines.clear();
        }
        let height = self.offset_to_line_col(text, text.len()).0 + 1;
        let last_line = min(last_line, height);

        // update find for given region
        if self.hls_dirty {
            self.update_find_for_lines(text, first_line, last_line);
        }

        let mut ops = Vec::new();
        if first_line > 0 {
            let op = if dirty { "invalidate" } else { "copy" };
            ops.push(self.build_update_op(op, None, first_line));
        }
        let first_line_offset = self.offset_of_line(text, first_line);
        let mut line_cursor = Cursor::new(text, first_line_offset);
        let mut soft_breaks = self.breaks.as_ref().map(|breaks|
            Cursor::new(breaks, first_line_offset)
        );

        let mut rendered_lines = Vec::new();
        for line_num in first_line..last_line {
            rendered_lines.push(self.render_line(tab_ctx, text,
                &mut line_cursor, soft_breaks.as_mut(), style_spans, line_num));
        }
        ops.push(self.build_update_op("ins", Some(rendered_lines), last_line - first_line));
        if last_line < height {
            if !dirty {
                ops.push(self.build_update_op("skip", None, last_line - first_line));
            }
            let op = if dirty { "invalidate" } else { "copy" };
            ops.push(self.build_update_op(op, None, height - last_line));
        }
        let params = json!({
            "ops": ops,
            "pristine": self.pristine,
        });
        tab_ctx.update_view(&self.view_id, &params);
        self.valid_lines.union_one_range(first_line, last_line);
    }


    /// Send lines within given region (plus slop) that the front-end does not already
    /// have.
    pub fn send_update_for_scroll<W: Write>(&mut self, text: &Rope, tab_ctx: &DocumentCtx<W>, style_spans: &Spans<Style>,
        first_line: usize, last_line: usize)
    {
        let first_line = max(first_line, SCROLL_SLOP) - SCROLL_SLOP;
        let last_line = last_line + SCROLL_SLOP;
        let height = self.offset_to_line_col(text, text.len()).0 + 1;
        let last_line = min(last_line, height);

        // update find for given region
        self.update_find_for_lines(text, first_line, last_line);

        let mut ops = Vec::new();
        let mut line = 0;
        for (start, end) in self.valid_lines.minus_one_range(first_line, last_line) {
            // TODO: this has some duplication with send_update in the non-dirty case.
            if start > line {
                ops.push(self.build_update_op("copy", None, start - line));
            }
            let start_offset = self.offset_of_line(text, start);
            let mut line_cursor = Cursor::new(text, start_offset);
            let mut soft_breaks = self.breaks.as_ref().map(|breaks|
                Cursor::new(breaks, start_offset)
            );
            let mut rendered_lines = Vec::new();
            for line_num in start..end {
                rendered_lines.push(self.render_line(tab_ctx, text,
                                                     &mut line_cursor, soft_breaks.as_mut(),
                                                     style_spans, line_num));
            }
            ops.push(self.build_update_op("ins", Some(rendered_lines), end - start));
            ops.push(self.build_update_op("skip", None, end - start));
            line = end;
        }
        if line == 0 {
            // Front-end already has all lines, no need to send any more.
            return;
        }
        if line < height {
            ops.push(self.build_update_op("copy", None, height - line));
        }
        let params = json!({
            "ops": ops,
            "pristine": self.pristine,
        });
        tab_ctx.update_view(&self.view_id, &params);
        self.valid_lines.union_one_range(first_line, last_line);
    }

    fn build_update_op(&self, op: &str, lines: Option<Vec<Value>>, n: usize) -> Value {
        let mut update = json!({
            "op": op,
            "n": n,
        });

        if let Some(lines) = lines {
            update["lines"] = json!(lines);
        }

        update
    }

    // Update front-end with any changes to view since the last time sent.
    pub fn render_if_dirty<W: Write>(&mut self, text: &Rope, tab_ctx: &DocumentCtx<W>, style_spans: &Spans<Style>) {
        if self.sel_dirty || self.hls_dirty || self.dirty {
            let first_line = max(self.first_line, SCROLL_SLOP) - SCROLL_SLOP;
            let last_line = self.first_line + self.height + SCROLL_SLOP;
            self.send_update(text, tab_ctx, style_spans, first_line, last_line);
            self.sel_dirty = false;
            self.hls_dirty = false;
            self.dirty = false;
        }
    }

    // TODO: finer grained tracking
    pub fn set_dirty(&mut self) {
        self.dirty = true;
    }

    // How should we count "column"? Valid choices include:
    // * Unicode codepoints
    // * grapheme clusters
    // * Unicode width (so CJK counts as 2)
    // * Actual measurement in text layout
    // * Code units in some encoding
    //
    // Of course, all these are identical for ASCII. For now we use UTF-8 code units
    // for simplicity.

    pub fn offset_to_line_col(&self, text: &Rope, offset: usize) -> (usize, usize) {
        let line = self.line_of_offset(text, offset);
        (line, offset - self.offset_of_line(text, line))
    }

    pub fn line_col_to_offset(&self, text: &Rope, line: usize, col: usize) -> usize {
        let mut offset = self.offset_of_line(text, line).saturating_add(col);
        if offset >= text.len() {
            offset = text.len();
            if self.line_of_offset(text, offset) <= line {
                return offset;
            }
        } else {
            // Snap to grapheme cluster boundary
            offset = text.prev_grapheme_offset(offset + 1).unwrap();
        }

        // clamp to end of line
        let next_line_offset = self.offset_of_line(text, line + 1);
        if offset >= next_line_offset {
            if let Some(prev) = text.prev_grapheme_offset(next_line_offset) {
                offset = prev;
            }
        }
        offset
    }

    // use own breaks if present, or text if not (no line wrapping)

    /// Returns the visible line number containing the given offset.
    pub fn line_of_offset(&self, text: &Rope, offset: usize) -> usize {
        match self.breaks {
            Some(ref breaks) => {
                breaks.convert_metrics::<BreaksBaseMetric, BreaksMetric>(offset)
            }
            None => text.line_of_offset(offset)
        }
    }

    /// Returns the byte offset corresponding to the line `line`.
    pub fn offset_of_line(&self, text: &Rope, line: usize) -> usize {
        match self.breaks {
            Some(ref breaks) => {
                breaks.convert_metrics::<BreaksMetric, BreaksBaseMetric>(line)
            }
            None => text.offset_of_line(line)
        }
    }

    pub fn rewrap(&mut self, text: &Rope, wrap_col: usize) {
        self.breaks = Some(linewrap::linewrap(text, wrap_col));
        self.wrap_col = wrap_col;
    }

    /// Updates the view after the text has been modified by the given `delta`.
    /// This method is responsible for updating the cursors, and also for
    /// recomputing line wraps.
    ///
    /// Return value is a location of a point that should be scrolled into view.
    pub fn after_edit(&mut self, text: &Rope, delta: &Delta<RopeInfo>, pristine: bool)
        -> Option<usize>
    {
        if let Some(breaks) = self.breaks.as_mut() {
            let (iv, new_len) = delta.summary();
            linewrap::rewrap(breaks, text, iv, new_len, self.wrap_col);
        }
        self.pristine = pristine;
        self.dirty = true;
        // Any edit cancels a drag. This is good behavior for edits initiated through
        // the front-end, but perhaps not for async edits.
        self.drag_state = None;

        // Update search highlights for changed regions
        // TODO: Incremental search instead of re-doing search completely
        if self.search_string.is_some() {
            self.valid_search.clear();
            self.occurrences = None;
            self.hls_dirty = true;
        }

        // Note: for committing plugin edits, we probably want to know the priority
        // of the delta so we can set the cursor before or after the edit, as needed.
        let new_sel = self.selection.apply_delta(delta, true);
        self.set_selection(text, new_sel)
    }

    /// Call to mark view as pristine. Used after a buffer is saved.
    pub fn set_pristine(&mut self) {
        self.pristine = true;
    }

    /// Unsets the search and removes all highlights from the view.
    pub fn unset_find(&mut self) {
        self.search_string = None;
        self.occurrences = None;
        self.hls_dirty = true;
        self.valid_search.clear();
    }

    /// Sets find for the view, highlights occurrences in the current viewport and selects the first
    /// occurrence relative to the last cursor.
    pub fn set_find(&mut self, search_string: &str, case_sensitive: bool) {
        let case_matching = if case_sensitive {
            CaseMatching::Exact
        } else {
            CaseMatching::CaseInsensitive
        };

        if let Some(ref s) = self.search_string {
            if s == search_string && case_matching == self.case_matching {
                // search parameters did not change
                return;
            }
        }

        self.unset_find();

        self.search_string = Some(search_string.to_string());
        self.case_matching = case_matching;
    }

    fn update_find_for_lines(&mut self, text: &Rope, first_line: usize, last_line: usize) {
        if self.search_string.is_none() {
            return;
        }
        let start = self.offset_of_line(text, first_line);
        let end = self.offset_of_line(text, last_line);
        self.update_find(text, start, end, true, false);
    }

    fn update_find(&mut self, text: &Rope, start: usize, end: usize, include_slop: bool,
                   stop_on_found: bool)
    {
        if self.search_string.is_none() {
            return;
        }

        // extend the search by twice the string length (twice, because case matching may increase
        // the length of an occurrence)
        let slop = if include_slop { self.search_string.as_ref().unwrap().len() * 2 } else { 0 };
        let mut occurrences = self.occurrences.take().unwrap_or_else(|| Selection::new());
        let mut searched_until = end;
        
        for (start, end) in self.valid_search.minus_one_range(start, end) {
            let search_string = self.search_string.as_ref().unwrap();
            let len = search_string.len();
        
            // expand region to be able to find occurrences around the region's edges
            let from = max(start, slop) - slop;
            let to = min(end + slop, text.len());

            let text = text.subseq(Interval::new_closed_open(0, to));
            let mut cursor = Cursor::new(&text, from);

            loop {
                match find(&mut cursor, self.case_matching, &search_string) {
                    Some(start) => {
                        let end = start + len;
                        let region = SelRegion {
                            start: start,
                            end: end,
                            horiz: None,
                            affinity: Affinity::default(),
                        };
                        occurrences.add_region(region);
                        
                        if stop_on_found {
                            searched_until = end;
                            break;
                        }
                    }
                    None => break,
                }
            }
        }
        self.occurrences = Some(occurrences);
        self.valid_search.union_one_range(start, searched_until);
        self.hls_dirty = true;
    }

    /// Select the next occurrence relative to the last cursor. `reverse` determines whether the
    /// next occurrence before (`true`) or after (`false`) the last cursor is selected. `wrapped`
    /// indicates a search for the next occurrence past the end of the file. `stop_on_found`
    /// determines whether the search should stop at the first found occurrence (does only apply
    /// to forward searc, i.e. reverse = false). If `allow_same` is set to `true` the current
    /// selection is considered a valid next occurrence.
    pub fn select_next_occurrence(&mut self, text: &Rope, reverse: bool, wrapped: bool,
                                  stop_on_found: bool, allow_same: bool) -> Option<usize>
    {
        if self.search_string.is_none() {
            return None;
        }

        let sel = match self.sel_regions().last() {
            Some(sel) => (sel.min(), sel.max()),
            None => return None
        };
        
        let (from, to) = if reverse != wrapped { (0, sel.0) } else { (sel.0, text.len()) };
        let mut next_occurrence;

        loop {
            next_occurrence = self.occurrences.as_ref().and_then(|occurrences| {
                if occurrences.len() == 0 {
                    return None;
                }
                if wrapped { // wrap around file boundaries 
                    if reverse {
                        occurrences.last()
                    } else {
                        occurrences.first()
                    }
                } else {
                    let ix = occurrences.search(sel.0);
                    if reverse {
                        ix.checked_sub(1).and_then(|i| occurrences.get(i))
                    } else {
                        occurrences.get(ix).and_then(|oc| {
                            // if possible, the current selection should be extended, instead of
                            // jumping to the next occurrence
                            if oc.end == sel.1 && !allow_same {
                                occurrences.get(ix+1)
                            } else {
                                Some(oc)
                            }
                        })
                    }
                }
            }).map(|o| o.clone());
            
            let region = {
                let mut unsearched = self.valid_search.minus_one_range(from, to);
                if reverse { unsearched.next_back() } else { unsearched.next() }
            };
            if let Some((b, e)) = region {
                if let Some(ref occurrence) = next_occurrence {
                    if (reverse && occurrence.start >= e) || (!reverse && occurrence.end <= b) {
                        break;
                    }
                }
                
                if !reverse {
                    self.update_find(text, b, e, false, stop_on_found);
                } else {
                    // when searching backward, the actual search isn't executed backwards, which is
                    // why the search is executed in chunks
                    let start = if e - b > BACKWARDS_FIND_CHUNK_SIZE {
                        e - BACKWARDS_FIND_CHUNK_SIZE
                    } else {
                        b
                    };
                    self.update_find(text, start, e, false, false);
                }
            } else {
                break;
            }
        }

        if let Some(occurrence) = next_occurrence {
            let mut selection = Selection::new();
            selection.add_region(occurrence);
            self.set_selection(text, selection)
        } else {
            None
        }
    }
}

// utility function to clamp a value within the given range
fn clamp(x: usize, min: usize, max: usize) -> usize {
    if x < min {
        min
    } else if x < max {
        x
    } else {
        max
    }
}
